from jax._src.numpy.vectorize import (
  functools,
  _apply_excluded,
  map,
  jnp,
  _parse_gufunc_signature,
  _parse_input_dimensions,
  _check_output_dims,
  zip,
  api,
)

def vectorize(pyfunc, *, excluded=frozenset(), signature=None):
  """Define a vectorized function with broadcasting.

  :func:`vectorize` is a convenience wrapper for defining vectorized
  functions with broadcasting, in the style of NumPy's
  `generalized universal functions <https://numpy.org/doc/stable/reference/c-api/generalized-ufuncs.html>`_.
  It allows for defining functions that are automatically repeated across
  any leading dimensions, without the implementation of the function needing to
  be concerned about how to handle higher dimensional inputs.

  :func:`jax.numpy.vectorize` has the same interface as
  :class:`numpy.vectorize`, but it is syntactic sugar for an auto-batching
  transformation (:func:`vmap`) rather than a Python loop. This should be
  considerably more efficient, but the implementation must be written in terms
  of functions that act on JAX arrays.

  Args:
    pyfunc: function to vectorize.
    excluded: optional set of integers representing positional arguments for
      which the function will not be vectorized. These will be passed directly
      to ``pyfunc`` unmodified.
    signature: optional generalized universal function signature, e.g.,
      ``(m,n),(n)->(m)`` for vectorized matrix-vector multiplication. If
      provided, ``pyfunc`` will be called with (and expected to return) arrays
      with shapes given by the size of corresponding core dimensions. By
      default, pyfunc is assumed to take scalars arrays as input and output.

  Returns:
    Vectorized version of the given function.

  Here are a few examples of how one could write vectorized linear algebra
  routines using :func:`vectorize`:

  >>> from functools import partial

  >>> @partial(jnp.vectorize, signature='(k),(k)->(k)')
  ... def cross_product(a, b):
  ...   assert a.shape == b.shape and a.ndim == b.ndim == 1
  ...   return jnp.array([a[1] * b[2] - a[2] * b[1],
  ...                     a[2] * b[0] - a[0] * b[2],
  ...                     a[0] * b[1] - a[1] * b[0]])

  >>> @partial(jnp.vectorize, signature='(n,m),(m)->(n)')
  ... def matrix_vector_product(matrix, vector):
  ...   assert matrix.ndim == 2 and matrix.shape[1:] == vector.shape
  ...   return matrix @ vector

  These functions are only written to handle 1D or 2D arrays (the ``assert``
  statements will never be violated), but with vectorize they support
  arbitrary dimensional inputs with NumPy style broadcasting, e.g.,

  >>> cross_product(jnp.ones(3), jnp.ones(3)).shape
  (3,)
  >>> cross_product(jnp.ones((2, 3)), jnp.ones(3)).shape
  (2, 3)
  >>> cross_product(jnp.ones((1, 2, 3)), jnp.ones((2, 1, 3))).shape
  (2, 2, 3)
  >>> matrix_vector_product(jnp.ones(3), jnp.ones(3))  # doctest: +IGNORE_EXCEPTION_DETAIL
  Traceback (most recent call last):
  ValueError: input with shape (3,) does not have enough dimensions for all
  core dimensions ('n', 'k') on vectorized function with excluded=frozenset()
  and signature='(n,k),(k)->(k)'
  >>> matrix_vector_product(jnp.ones((2, 3)), jnp.ones(3)).shape
  (2,)
  >>> matrix_vector_product(jnp.ones((2, 3)), jnp.ones((4, 3))).shape
  (4, 2)

  Note that this has different semantics than `jnp.matmul`:

  >>> jnp.matmul(jnp.ones((2, 3)), jnp.ones((4, 3)))  # doctest: +IGNORE_EXCEPTION_DETAIL
  Traceback (most recent call last):
  TypeError: dot_general requires contracting dimensions to have the same shape, got [3] and [4].
  """
  if any(not isinstance(exclude, int) for exclude in excluded):
    raise TypeError("jax.numpy.vectorize can only exclude integer arguments, "
                    "but excluded={!r}".format(excluded))
  if excluded and min(excluded) < 0:
    raise ValueError(f"excluded={excluded!r} contains negative numbers")

  @functools.wraps(pyfunc)
  def wrapped(*args):
    error_context = ("on vectorized function with excluded={!r} and "
                     "signature={!r}".format(excluded, signature))
    excluded_func, args = _apply_excluded(pyfunc, excluded, args)
    args = tuple(map(jnp.asarray, args))

    if signature is not None:
      input_core_dims, output_core_dims = _parse_gufunc_signature(signature)
    else:
      input_core_dims = [()] * len(args)
      output_core_dims = None

    broadcast_shape, dim_sizes = _parse_input_dimensions(
        args, input_core_dims, error_context)

    checked_func = _check_output_dims(
        excluded_func, dim_sizes, output_core_dims, error_context)

    # Rather than broadcasting all arguments to full broadcast shapes, prefer
    # expanding dimensions using vmap. By pushing broadcasting
    # into vmap, we can make use of more efficient batching rules for
    # primitives where only some arguments are batched (e.g., for
    # lax_linalg.triangular_solve), and avoid instantiating large broadcasted
    # arrays.

    squeezed_args = []
    rev_filled_shapes = []

    for arg, core_dims in zip(args, input_core_dims):
      noncore_shape = arg.shape[:arg.ndim - len(core_dims)]
      
      pad_ndim = len(broadcast_shape) - len(noncore_shape)
      filled_shape = pad_ndim * (1,) + noncore_shape
      assert len(filled_shape) == len(broadcast_shape)
      assert jnp.broadcast_shapes(broadcast_shape, filled_shape) == broadcast_shape
      rev_filled_shapes.append(filled_shape[::-1])

      squeeze_indices = tuple(i for i, size in enumerate(noncore_shape) if size == 1)
      squeezed_arg = jnp.squeeze(arg, axis=squeeze_indices)
      squeezed_args.append(squeezed_arg)

    vectorized_func = checked_func
    for axis_sizes in zip(*rev_filled_shapes):
      in_axes = tuple(None if size == 1 else 0 for size in axis_sizes)
      vectorized_func = api.vmap(vectorized_func, in_axes)
    return vectorized_func(*squeezed_args)

  return wrapped



class SingletonMeta(type):
    
    def __repr__(cls):
        return cls.__name__

class Singleton(metaclass=SingletonMeta):
    
    def __new__(cls):
        raise NotImplementedError(f"{cls.__name__} can not be instantiated")

class Cippa(Singleton):
    """cippa"""
    pass



import lsqfitgp as lgp
from matplotlib import pyplot as plt
import numpy as np

"""Check the symmetry of the Bernoulli polynomials"""

fig, axs = plt.subplots(1, 2, num='bernoulli', clear=True, figsize=[8, 5])

axs[0].set_title('$B_n(x)$')
axs[1].set_title('$B_n(x) - (-1)^n B_n(1-x)$')

eps = np.finfo(float).eps
x = np.linspace(eps, 1 - eps, 1000)

for n in range(2, 11):
    y = lgp._patch_jax._periodic_bernoulli(n, x)
    yrev = (-1) ** n * lgp._patch_jax._periodic_bernoulli(n, 1 - x)
    axs[0].plot(x, y, label=str(n))
    axs[1].plot(x, y - yrev, label=str(n))

for ax in axs:
    ax.legend()

fig.tight_layout()
fig.show()



# lsqfitgp/copula/_invgamma.py
#
# Copyright (c) 2023, Giacomo Petrillo
#
# This file is part of lsqfitgp.
#
# lsqfitgp is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# lsqfitgp is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with lsqfitgp.  If not, see <http://www.gnu.org/licenses/>.

"""JAX-compatible implementation of the inverse gamma distribution"""

import functools

from scipy import special
import jax
from jax.scipy import special as jspecial
from jax import numpy as jnp
import numpy

from .. import _patch_jax

def _castto(func, type):
    @functools.wraps(func)
    def newfunc(*args, **kw):
        return func(*args, **kw).astype(type)
    return newfunc

@jax.custom_jvp
def gammainccinv(a, y):
    a = jnp.asarray(a)
    y = jnp.asarray(y)
    class ResultDummy:
        dtype = jnp.result_type(a.dtype, y.dtype)
        shape = jnp.broadcast_shapes(a.shape, y.shape)
    ufunc = _castto(special.gammainccinv, ResultDummy.dtype)
    return jax.pure_callback(ufunc, ResultDummy, a, y, vectorized=True)

dQ_da = _patch_jax.elementwise_grad(jspecial.gammaincc, 0)
dQ_dx = _patch_jax.elementwise_grad(jspecial.gammaincc, 1)

@gammainccinv.defjvp
def gammainccinv_jvp(primals, tangents):
    a, y = primals
    at, yt = tangents
    x = gammainccinv(a, y)
    a = jnp.asarray(a)
    a = a.astype(jnp.sin(jnp.empty(0, a.dtype)).dtype)
    # convert a to float to avoid dQ_da complaining even when we are not
    # actually deriving w.r.t. a
    dQ_da_a_x = dQ_da(a, x)
    dQ_dx_a_x = dQ_dx(a, x)
    dQinv_dy_a_y = 1 / dQ_dx_a_x
    dQinv_da_a_y = -dQinv_dy_a_y * dQ_da_a_x
    return x, dQinv_da_a_y * at + dQinv_dy_a_y * yt

@jax.custom_jvp
def gammaincinv(a, y):
    a = jnp.asarray(a)
    y = jnp.asarray(y)
    class ResultDummy:
        dtype = jnp.result_type(a.dtype, y.dtype)
        shape = jnp.broadcast_shapes(a.shape, y.shape)
    ufunc = _castto(special.gammaincinv, ResultDummy.dtype)
    return jax.pure_callback(ufunc, ResultDummy, a, y, vectorized=True)

dP_da = _patch_jax.elementwise_grad(jspecial.gammainc, 0)
dP_dx = _patch_jax.elementwise_grad(jspecial.gammainc, 1)

@gammaincinv.defjvp
def gammaincinv_jvp(primals, tangents):
    a, y = primals
    at, yt = tangents
    x = gammaincinv(a, y)
    a = jnp.asarray(a)
    a = a.astype(jnp.sin(jnp.empty(0, a.dtype)).dtype)
    # convert a to float to avoid dQ_da complaining even when we are not
    # actually deriving w.r.t. a
    dP_da_a_x = dP_da(a, x)
    dP_dx_a_x = dP_dx(a, x)
    dPinv_dy_a_y = 1 / dP_dx_a_x
    dPinv_da_a_y = -dPinv_dy_a_y * dP_da_a_x
    return x, dPinv_da_a_y * at + dPinv_dy_a_y * yt

class invgamma:
    
    @staticmethod
    def ppf(q, a, scale=1):
        return scale / gammainccinv(a, q)

    @staticmethod
    def isf(q, a, scale=1):
        return scale / gammaincinv(a, q)

    @staticmethod
    def logpdf(x, a, scale=1):
        x = x / scale
        return -jnp.log(scale) - (a + 1) * jnp.log(x) - 1 / x - jspecial.gammaln(a)

    @staticmethod
    def cdf(x, a, scale=1):
        return jspecial.gammaincc(a, scale / x)



import numpy as np
from matplotlib import pyplot as plt

nd = 2
alpha = 1
beta = np.ones(nd)
nx = 20

def kernel(x, y, alpha, beta):
    bn2 = np.sum(beta ** 2)
    xyb = x + y + beta
    xybn2 = np.sum(xyb ** 2, axis=-1)
    return (bn2 / xybn2) ** (alpha / 2)

gen = np.random.default_rng(202206241555)
x = gen.uniform(0, 5, (nx, nd))
c = kernel(x[None, :, :], x[:, None, :], alpha, beta)
w = np.linalg.eigvalsh(c)
error = len(c) * np.finfo(float).eps * np.max(w)
if np.min(w) < -error:
    print('NOT POSITIVE!')

fig, ax = plt.subplots(num='decaying', clear=True)

ax.plot(w)

fig.show()



# lsqfitgp/_kernels/__init__.py
#
# Copyright (c) 2023, Giacomo Petrillo
#
# This file is part of lsqfitgp.
#
# lsqfitgp is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# lsqfitgp is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with lsqfitgp.  If not, see <http://www.gnu.org/licenses/>.

from ._arma import *
from ._bart import *
from ._basic import *
from ._celerite import *
from ._matern import *
from ._randomwalk import *
from ._spectral import *
from ._wendland import *
from ._zeta import *

# TODO instead of adding forcekron by default to all 1D kernels, use maxdim=None
# by default in CrossKernel, add maxdim=1 to all 1D kernels, and let the user
# choose how to deal with nd (add option for sum-separation). Make an example
# about this in `multidimensional input`. Implement tests for separability
# on all kernels.

# TODO maybe I could have a continuity check like derivable, but to be useful
# it would be callable-only and take the derivation order. But I don't think
# any potential user needs it.

# TODO add explicit exponent parameter to infinitely divisible kernels, and
# check the exponent is an int in __pow__.



import pymc3 as pm
import numpy as np ####
import gvar ####
from scipy import stats ####

x = np.linspace(-5, 5, 11) ####
y = np.sin(x) ####

model = pm.Model()
with model:
    logscale = pm.Normal('logscale', mu=np.log(3), sigma=1)
    logsdev = pm.Normal('logsdev', mu=np.log(1), sigma=1)
    kernel = pm.math.exp(logsdev) ** 2 * pm.gp.cov.ExpQuad(1, ls=pm.math.exp(logscale))
    gp = pm.gp.Marginal(cov_func=kernel)
    gp.marginal_likelihood('data', x[:, None], y, 0)

with model:
    mp = pm.find_MAP()
    trace = pm.sample(10000, cores=1)

print('\nMaximum a posteriori (must be the same as lsqfitgp):')
print('log(sdev) {:.2f}'.format(mp['logsdev']))
print('log(scale) {:.2f}'.format(mp['logscale']))

df = pm.trace_to_dataframe(trace)
mean = df.mean()
cov = df.cov()

meandict = {}
covdict = {}
for label1 in df.columns:
    meandict[label1] = mean[label1]
    for label2 in df.columns:
        covdict[label1, label2] = cov[label1][label2]

params = gvar.gvar(meandict, covdict)
print('\nPosterior mean and standard deviation:')
print('log(sdev)', params['logsdev'])
print('log(scale)', params['logscale'])

p = params['logsdev']
prob_gauss = stats.norm.cdf(np.log(1), loc=gvar.mean(p), scale=gvar.sdev(p))
true_prob = np.sum(df['logsdev'] <= np.log(1)) / len(df)
print('\nProbability of having sdev < 1:')
print('prob_gauss {:.3g}'.format(prob_gauss))
print('true_prob {:.3g}'.format(true_prob))



import functools

class lippa:
    
    def __init__(self):
        self.impl = lambda x: x
    
    def __call__(self, arg):
        return self.impl(arg)

    def lupi(self, num):
        orig = self.impl
        transf = self.lupi.impls[num](self)
        new = lippa()
        new.impl = lambda x: transf(orig(x))
        return new
    
def impl(meth):
    oldmeth = getattr(lippa, meth.__name__)
    @functools.wraps(oldmeth)
    def newmeth(*args):
        return oldmeth(*args)
    newmeth.impls = {0: meth}
    def one(meth):
        newmeth.impls[1] = meth
        return newmeth
    newmeth.one = one
    return newmeth
    
class cippa(lippa):
    
    @impl
    def lupi(self):
        return lambda x: x
    
    @lupi.one
    def lupi(self):
        return lambda x: x ** 2

u = cippa()

print(cippa.lupi.impls[0](u)(2)) # -> 2
print(cippa.lupi.impls[1](u)(2)) # -> 4

print(u.lupi.impls[0](u)(2)) # -> 2
print(u.lupi.impls[1](u)(2)) # -> 4

print(u.lupi(0)(2)) # -> 2
print(u.lupi(1)(2)) # -> 4



"""Fit of parton distributions functions (PDFs)

Like pdf3, but with nonlinear data"""

import lsqfitgp as lgp
import numpy as np
from matplotlib import pyplot as plt
import gvar
import lsqfit

np.random.seed(20220416)

#### COMPONENTS ####

flavor = np.array([
    ( 1, 'd'    ), # 0
    (-1, 'dbar' ), # 1
    ( 2, 'u'    ), # 2
    (-2, 'ubar' ), # 3
    ( 3, 's'    ), # 4
    (-3, 'sbar' ), # 5
    ( 4, 'c'    ), # 6
    (-4, 'cbar' ), # 7
    (21, 'gluon'), # 8
], 'i8, U16')

pid  = flavor['f0']
name = flavor['f1']

nflav  = len(flavor)
nx     = 30
ndata  = 10
nx2    = 30  # must be <= nx
ndata2 = 10

indices = dict(
    # quark, antiquark
    d = [0, 1],
    u = [2, 3],
    s = [4, 5],
    c = [6, 7],
)

#### MODEL ####
# for each component of the proton:
# h ~ GP
# f = h''  (f is the PDF)
# for the momentum sum rule:
# int_0^1 dx x f(x) = [xh'(x) - h(x)]_0^1
# for the flavor sum rules:
# int_0^1 dx (f_i(x) - f_j(x)) = [h_i'(x) - h_j'(x)]_0^1

xtype = np.dtype([
    ('x'  , float),
    ('pid', int  ),
])

kernel = lgp.ExpQuad(dim='x') * lgp.White(dim='pid')

xdata = np.empty((nflav, nx), xtype)
xdata['pid'] = pid[:, None]
xdata[  'x'] = np.linspace(0, 1, nx)

M = np.random.randn(ndata, nflav, nx) # linear map PDF(X) -> data

A = np.random.randn(ndata2, nflav, nx2, nx2)
M2 = np.einsum('dfxz,dfyz->dfxy', A, A) # quadratic map PDF(X) -> data2

xinteg = np.empty((nflav, 2), xtype)
xinteg['pid'] = pid[:, None]
xinteg[  'x'] = [0, 1]

suminteg = np.empty(xinteg.shape)
suminteg[:, 0] = -1
suminteg[:, 1] =  1

constraints = {
    'momrule': 1,
    'uubar'  : 2,
    'ddbar'  : 1,
    'ccbar'  : 0,
    'ssbar'  : 0,
}

#### CREATE GP OBJECT ####

gp = lgp.GP()

gp.addproc(kernel, 'h')
gp.addproctransf({'h': 1}, 'primitive', deriv='x'     )
gp.addproctransf({'h': 1}, 'f'        , deriv=(2, 'x'))
gp.addproctransf({
    'primitive': lambda x: x['x'],
    'h'        : -1,
}, 'primitive of xf(x)')

# data
gp.addx(xdata, 'xdata', proc='f')
gp.addtransf({'xdata': M}, 'data', axes=2)

# total momentum rule
gp.addx(xinteg, 'xmomrule', proc='primitive of xf(x)')
gp.addtransf({'xmomrule': suminteg}, 'momrule', axes=2)

# quark sum rules
qdiff = np.array([1, -1])[:, None]
for quark in 'ducs':
    idx = indices[quark]
    label = f'{quark}{quark}bar' # the one appearing in `constraints`
    xlabel = f'x{label}'
    gp.addx(xinteg[idx], xlabel, proc='primitive')
    gp.addtransf({xlabel: suminteg[idx] * qdiff}, label, axes=2)
    
#### DEFINE NONLINEAR FUNCTION ####

def fcn(params):
    
    xdata = params['xdata']
    # data2 = np.einsum('dfxy,fx,fy->d', M2, xdata, xdata)
    xdata2 = xdata[:, None, :nx2] * xdata[:, :nx2, None]
    data2 = np.tensordot(M2, xdata2, axes=3)
    
    return dict(data=params['data'], data2=data2)

params_prior = gp.predfromdata(constraints, ['xdata', 'data'])

#### GENERATE FAKE DATA ####

priorsample = next(gvar.raniter(params_prior))
fcnsample = fcn(priorsample)

datamean = gvar.BufferDict({
    'data' : fcnsample['data' ],
    'data2': fcnsample['data2'],
})
dataerr = gvar.BufferDict({
    'data' : np.full(ndata ,   1),
    'data2': np.full(ndata2, 100),
})
datamean.buf += dataerr.buf * np.random.randn(*dataerr.buf.shape)
data = gvar.gvar(datamean, dataerr)

# check sum rules approximately with trapezoid rule
def check_integrals(x, y):
    checksum = np.sum(((y * x)[:, 1:] + (y * x)[:, :-1]) / 2 * np.diff(x, axis=1))
    print('sum_i int dx x f_i(x) =', checksum)
    for q in 'ducs':
        idx = indices[q]
        qx = x[idx]
        qy = y[idx]
        checksum = np.sum(qdiff * (qy[:, 1:] + qy[:, :-1]) / 2 * np.diff(qx, axis=1))
        print(f'sum_i={q}{q}bar int dx f_i(x) =', checksum)

print('check integrals in fake data:')
check_integrals(xdata['x'], priorsample['xdata'])

#### FIT ####

fit = lsqfit.nonlinear_fit(data, fcn, params_prior, verbose=2)
print(fit.format(maxline=True))
print(fit.format(maxline=-1))

pred = fcn(fit.p)

print('check integrals in fit:')
check_integrals(xdata['x'], fit.p['xdata'])

#### PLOT RESULTS ####

fig, axs = plt.subplots(1, 3, num='pdf5', clear=True, figsize=[12, 4.5])
axs[0].set_title('PDFs')
axs[1].set_title('Data')
axs[2].set_title('Data (quadratic)')

for i in range(nflav):
    
    x = xdata[i]['x']
    ypdf = fit.p['xdata'][i]
    ydata = priorsample['xdata'][i]
    m = gvar.mean(ypdf)
    s = gvar.sdev(ypdf)
    
    color = 'C' + str(i // 2)
    if i % 2:
        kw = dict(hatch='//////', edgecolor=color, facecolor='none')
        kwp = dict(color=color, linestyle='--')
    else:
        kw = dict(alpha=0.6, facecolor=color)
        kwp = dict(color=color)
    
    axs[0].fill_between(x, m - s, m + s, label=name[i], **kw)
    axs[0].plot(x, ydata, **kwp)

axs[0].legend(fontsize='small')

for ax, label in zip(axs[1:], ['data', 'data2']):
    
    m = gvar.mean(pred[label])
    s = gvar.sdev(pred[label]) if len(m) else []
    x = np.arange(len(m))
    ax.fill_between(x, m - s, m + s, step='mid', color='lightgray')
    ax.errorbar(x, datamean[label], dataerr[label], color='black', linestyle='', capsize=2)
    ax.plot(x, fcnsample[label], drawstyle='steps-mid', color='black')

fig.tight_layout()
fig.show()



# lsqfitgp/_patch_gvar.py
#
# Copyright (c) 2020, 2022, 2023, Giacomo Petrillo
#
# This file is part of lsqfitgp.
#
# lsqfitgp is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# lsqfitgp is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with lsqfitgp.  If not, see <http://www.gnu.org/licenses/>.

import functools
import textwrap

import jax
import gvar
from jax import numpy as jnp
from jax.scipy import special as jspecial
from jax import tree_util
import numpy
from scipy import linalg

from . import _patch_jax

gvar_ufuncs = [
    'sin',
    'cos',
    'tan',
    'exp',
    'log',
    'sqrt',
    'fabs',
    'sinh',
    'cosh',
    'tanh',
    'arcsin',
    'arccos',
    'arctan',
    'arctan2',
    'arcsinh',
    'arccosh',
    'arctanh',
    'square',
    'erf',
]

for fname in gvar_ufuncs:
    fgvar = getattr(gvar, fname)
    fjax = getattr(jnp, fname, getattr(jspecial, fname, NotImplemented))
    fboth = functools.singledispatch(fgvar)
    fboth.register(jnp.ndarray, fjax)
    setattr(gvar, fname, fboth)

# reset transformations to support jax arrays
gvar.BufferDict.del_distribution('log')
gvar.BufferDict.del_distribution('sqrt')
gvar.BufferDict.del_distribution('erfinv')
gvar.BufferDict.add_distribution('log', gvar.exp)
gvar.BufferDict.add_distribution('sqrt', gvar.square)
gvar.BufferDict.add_distribution('erfinv', gvar.erf)

def _getsvec(x):
    """
    Get the sparse vector of derivatives of a GVar.
    """
    if isinstance(x, gvar.GVar):
        return x.internaldata[1]
    else:
        return gvar.svec(0)

def _merge_svec(gvlist, start=None, stop=None):
    if start is None:
        return _merge_svec(gvlist, 0, len(gvlist))
    n = stop - start
    if n <= 0:
        return gvar.svec(0)
    if n == 1:
        return _getsvec(gvlist[start])
    left = _merge_svec(gvlist, start, start + n // 2)
    right = _merge_svec(gvlist, start + n // 2, stop)
    return left.add(right, 1, 1)

def jacobian(g):
    """
    Extract the jacobian of gvars w.r.t. primary gvars.
    
    Parameters
    ----------
    g : array_like
        An array of numbers or gvars.
    
    Return
    ------
    jac : array
        The shape is g.shape + (m,), where m is the total number of primary
        gvars that g depends on.
    indices : (m,) int array
        The indices that map the last axis of jac to primary gvars in the
        global covariance matrix.
    """
    g = numpy.asarray(g)
    v = _merge_svec(g.flat)
    indices = v.indices()
    jac = numpy.zeros((g.size, len(indices)), float)
    for i, x in enumerate(g.flat):
        v = _getsvec(x)
        ind = numpy.searchsorted(indices, v.indices())
        jac[i, ind] = v.values()
    jac = jac.reshape(g.shape + indices.shape)
    return jac, indices

def from_jacobian(mean, jac, indices):
    """
    Create new gvars from a jacobian w.r.t. primary gvars.
    
    Parameters
    ----------
    mean : array_like
        An array of numbers with the means of the new gvars.
    jac : mean.shape + (m,) array
        The derivatives of each new gvar w.r.t. m primary gvars.
    indices : (m,) int array
        The indices of the primary gvars.
    
    Return
    ------
    g : mean.shape array
        The new gvars.
    """
    cov = gvar.gvar.cov
    mean = numpy.asarray(mean)
    shape = mean.shape
    mean = mean.flat
    jac = numpy.asarray(jac)
    jac = jac.reshape(len(mean), len(indices))
    g = numpy.zeros(len(mean), object)
    for i, jacrow in enumerate(jac):
        der = gvar.svec(len(indices))
        der._assign(jacrow, indices)
        g[i] = gvar.GVar(mean[i], der, cov)
    return g.reshape(shape)

def _skeleton(bd):
    return gvar.BufferDict(
        bd,
        buf=numpy.broadcast_to(numpy.empty((), bd.dtype), bd.buf.shape),
        # buf is not copied, so the memoryless broadcast is kept
    )

def bufferdict_flatten(bd):
    return (bd.buf,), _skeleton(bd)

def bufferdict_unflatten(skeleton, children):
    buf, = children
    new = _skeleton(skeleton)
    # copy the skeleton to permit multiple unflattening
    new._extension = {}
    new._buf = buf
    return new

# register BufferDict as a pytree
tree_util.register_pytree_node(gvar.BufferDict, bufferdict_flatten, bufferdict_unflatten)

# TODO the current implementation of BufferDict as pytree is not really
# consistent with how JAX handles trees, because JAX expects to be allowed to
# put arbitrary objects in the leaves; in particular, internally it sometimes
# creates dummy trees filled with None. Maybe the current impl is fine with
# this; _buf gets set to None, and assuming the BufferDict is never really
# used in that crooked state, everything goes fine. The thing that this breaks
# is a subsequent flattening of the dummy, I think JAX never does this. (The
# reason for switching to buf-as-leaf in place of dict-values-as-leaves is that
# the latter breaks tracing.)

def tabulate_together(*gs, headers=True, offset='', ndecimal=None, keys=None):
    """
    
    Format a table comparing side by side various collections of gvars.
    
    Parameters
    ----------
    *gs : sequence of arrays or dictionaries of gvars
        The variables to be tabulated. The structures of arrays and dictionaries
        must match.
    headers : bool or sequence of strings
        If True (default), add automatically an header. If False, don't add an
        header. If a sequence with length len(gs) + 1, it contains the column
        names for the keys/indices and for each set of variables.
    offset : str
        Prefix to each line, default empty.
    ndecimal : int, optional
        Number of decimal places. If not specified (default), keep two error
        digits.
    keys : sequence, optional
        If `gs` are dictionaries, a subset of keys to be extracted from each
        dictionary. Ignored if they are arrays.
    
    Examples
    --------
    >>> print(tabulate_together(gvar.gvar(dict(a=1)), gvar.gvar(dict(a=2))))
    key/index   value1   value2
    ---------------------------
            a    1 (0)    2 (0)
    
    See also
    --------
    gvar.tabulate
    
    """
    if not gs:
        return ''
    gs = [g if hasattr(g, 'keys') else numpy.asarray(g) for g in gs]
    assert all(hasattr(g, 'keys') for g in gs) or all(not hasattr(g, 'keys') for g in gs)
    if keys is not None and hasattr(gs[0], 'keys'):
        gs = [{k: g[k] for k in keys} for g in gs]
    g0 = gs[0]
    if hasattr(g0, 'keys'):
        assert all(set(g.keys()) == set(g0.keys()) for g in gs[1:])
        gs = [{k: g[k] for k in g0} for g in gs]
    else:
        assert all(g.shape == g0.shape for g in gs[1:])
        if g0.shape == ():
            gs = [{'--': g} for g in gs]
    tables = [
        _splittable(gvar.tabulate(g, headers=['@', ''], ndecimal=ndecimal))
        for g in gs
    ]
    columns = list(tables[0]) + [t[1] for t in tables[1:]]
    if not hasattr(headers, '__len__'):
        if headers:
            headers = ['key/index'] + [f'value{i+1}' for i in range(len(gs))]
        else:
            headers = None
    else:
        assert len(headers) == len(columns)
    if headers is not None:
        columns = (_head(col, head) for col, head in zip(columns, headers))
    return textwrap.indent(_join(columns), offset)

def _splittable(table):
    lines = table.split('\n')
    header = lines[0]
    col = header.find('@') + 1
    contentlines = lines[2:]
    col1 = '\n'.join(line[:col] for line in contentlines)
    col2 = '\n'.join(line[col:] for line in contentlines)
    return col1, col2

def _head(col, head):
    head = str(head)
    width = col.find('\n')
    if width < 0:
        width = len(col)
    hwidth = len(head)
    if hwidth > width:
        col = textwrap.indent(col, (hwidth - width) * ' ')
    else:
        head = (width - hwidth) * ' ' + head
    return head + '\n' + len(head) * '-' + '\n' + col

def _join(cols):
    split = (col.split('\n') for col in cols)
    return '\n'.join(''.join(lines) for lines in zip(*split))

def add_gvar_support(func):
    """ Wraps a jax-traceable ufunc to support gvars """
    
    dfdx = _patch_jax.elementwise_grad(func)
    
    def gvar_function(x):
        m = gvar.mean(x)
        return gvar.gvar_function(x, func(m), dfdx(m))
    
    gvar_function_vectorized = numpy.vectorize(gvar_function)

    @functools.wraps(func)
    def decorated_func(x):
        if isinstance(x, gvar.GVar):
            return gvar_function(x)
        elif getattr(x, 'dtype', None) == object:
            return gvar_function_vectorized(x)
        else:
            return func(x)
    
    return decorated_func



# lsqfitgp/examples/k.py
#
# Copyright (c) 2020, 2022, 2023, Giacomo Petrillo
#
# This file is part of lsqfitgp.
#
# lsqfitgp is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# lsqfitgp is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with lsqfitgp.  If not, see <http://www.gnu.org/licenses/>.

"""

                            EXAMPLE K.

    Where lady K finds out that every gentleman hides his rougher
    corners in his derivative.

"""

import lsqfitgp as lgp
from matplotlib import pyplot as plt
import numpy as np
import gvar

xdata = np.linspace(0, 10, 10)
xpred = np.linspace(-15, 25, 200)
y = np.sin(xdata)

print('make GP...')
gp = lgp.GP(lgp.Maternp(p=2, scale=3), checkpos=False)
gp.addx(xdata, 'data')
gp.addx(xpred, 'pred')
gp.addx(xpred, 'deriv', deriv=1)

print('fit...')
u = gp.pred({'data': y}, ['pred', 'deriv'], fromdata=True)

print('figure...')
fig, ax = plt.subplots(num='k', clear=True)

colors = dict()
for label in u:
    m = gvar.mean(u[label])
    s = gvar.sdev(u[label])
    patch = ax.fill_between(xpred, m - s, m + s, label=label, alpha=0.5)
    colors[label] = patch.get_facecolor()[0]
    
print('samples...')
for sample in gvar.raniter(u, 1):
    for label in u:
        ax.plot(xpred, sample[label], '-', color=colors[label])
ax.plot(xdata, y, 'k.', label='data')
ax.legend(loc='best')

fig.show()



# lsqfitgp/docs/kernelsref.py
#
# Copyright (c) 2020, 2022, 2023, Giacomo Petrillo
#
# This file is part of lsqfitgp.
#
# lsqfitgp is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# lsqfitgp is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with lsqfitgp.  If not, see <http://www.gnu.org/licenses/>.

""" Generate a file with the list of kernels """

outputfile = 'kernelsref.rst'

import inspect
import sys

import numpy as np
from matplotlib import pyplot as plt

sys.path = ['.', '..'] + sys.path
import lsqfitgp as lgp

classes = (lgp.IsotropicKernel, lgp.StationaryKernel, lgp.Kernel     )
titles  = ('Isotropic kernels', 'Stationary kernels', 'Other kernels')

kernels = []
for name, obj in vars(lgp).items():
    if inspect.isclass(obj) and issubclass(obj, lgp.Kernel):
        if obj not in classes:
            kernels.append(name)
kernels.sort()

out = """\
.. file generated automatically by lsqfitgp/docs/kernelsref.py

.. currentmodule:: lsqfitgp

.. _kernels:

Kernels reference
=================

This is a list of all the specific kernels implemented in :mod:`lsqfitgp`.

Kernels are reported with a simplified signature where the positional arguments
are `r` or `r2` if the kernel is isotropic, `delta` if it is stationary, or
`x`, `y` for generic kernels, and with only the keyword arguments specific to
the kernel. All kernels also understand the general keyword arguments of
:class:`Kernel` (or their specific superclass), while there are no positional
arguments when instantiating the kernel and the call signature of instances is
always `x`, `y`.

Example: the kernel :class:`GammaExp` is listed as ``GammaExp(r, gamma=1)``.
This means you could use it this way::

    import lsqfitgp as lgp
    import numpy as np
    kernel = lgp.GammaExp(loc=0.3, scale=2, gamma=1.4)
    x = np.random.randn(100)
    covmat = kernel(x[:, None], x[None, :])

On multidimensional input, isotropic kernels will compute the euclidean
distance. In general non-isotropic kernels will act separately on each
dimension, i.e., :math:`k(x_1,y_1,x_2,y_2) = k(x_1,y_1) k(x_2,y_2)`, apart from
kernels defined in terms of the dot product.

For all isotropic and stationary (i.e., depending only on :math:`x - y`)
kernels :math:`k(x, x) = 1`, and the typical lengthscale is approximately 1 for
default values of the keyword parameters, apart from some specific cases like
:class:`Constant`.

.. warning::

   You may encounter problems with second derivatives for
   :class:`CausalExpQuad`, :class:`FracBrownian`, :class:`NNKernel`, and with
   first derivatives too for :class:`Wendland` (but only in more than one
   dimension). :class:`Color` stops working for :math:`n > 20`.

Index
-----
"""

# index of kernels
kernels2 = list(kernels)
kernels = kernels[::-1]
for superclass, title in zip(classes, titles):
    out += f"""
{title}
{'^' * len(title)}
"""
    for i in reversed(range(len(kernels))):
        obj = getattr(lgp, kernels[i])
        if issubclass(obj, superclass):
            out += f"""\
  * :class:`{kernels[i]}`
"""
            del kernels[i]

out += """
Documentation
-------------
"""

class Formula:
    
    def __init__(self, formula):
        self.formula = formula
    
    def __repr__(self):
        return self.formula
    
    def __call__(self, x):
        return eval(self.formula, vars(np), dict(x=x))

# TODO nd plot for kernels with maxdim > 1, key kwlist_nd has precedence over
# kwlist

def bart_splits(n):
    x = np.random.uniform(1, 4, (n, 1))
    l, s = lgp.BART.splits_from_coord(x)
    return [l.item()], np.asarray(s.squeeze(1))

meta = dict(
    AR = dict(x=np.arange(50), kwlist=[
        dict(phi=np.array(phi), maxlag=50, norm=True) for phi in [
            [0.9], [1.82, -0.83],
        ]
    ] + [
        dict(gamma=np.array(gamma), maxlag=50) for gamma in [
            [1, 0.99],
        ]
    ] + [
        dict(slnr=np.array(r), lnc=np.array(c), norm=True) for r, c in [
            ([0.1], [0.1 + 1j]),
            (3 * [0.2], []),
        ]
    ]),
    BART = dict(kwlist=[dict(splits=bart_splits(100), maxd=d) for d in [2, 1]]),
    BagOfWords = dict(skip=True),
    Bessel = dict(range=[0, 10], kwlist=[dict(nu=v) for v in [0, 1, 2, 3.5]]),
    BrownianBridge = dict(range=[0, 1]),
    Categorical = dict(skip=True),
    Cauchy = dict(kwlist=[dict(alpha=1), dict(alpha=2), dict(beta=10)]),
    CausalExpQuad = dict(kwlist=[dict(alpha=a) for a in [0, 1, 2]]),
    Celerite = dict(kwlist=[dict(B=0), dict(B=1), dict(gamma=5)]),
    Circular = dict(kwlist=[dict(c=c, tau=t) for c, t in [(1/2, 4), (1/2, 10), (1/4, 4)]], range=[0, 2]),
    Color = dict(range=[0, 4 * np.pi], kwlist=[dict(n=n) for n in [2, 4, 6, 20]]),
    Constant = dict(skip=True),
    Cos = dict(range=[0, 4 * np.pi]),
    Decaying = dict(range=[0, 2], srange=[0, 5]),
    FracBrownian = dict(kwlist=[dict(H=H, K=K) for H, K in [(0.1, 1), (0.5, 1), (0.9, 1), (0.9, 0.3)]], range=[-5, 5]),
    GammaExp = dict(kwlist=[dict(gamma=g) for g in [0.1, 1, 1.9]]),
    Gibbs = dict(kwlist=[dict(scalefun=Formula('where((0 < x) & (x < 0.1), 0.02, 1)'))], range=[-1, 1]),
    Harmonic = dict(range=[0, 4 * np.pi], kwlist=[dict(Q=Q) for Q in [1/20, 1, 20]]),
    MA = dict(x=np.arange(50), kwlist=[dict(w=np.array(w)) for w in [
        2 * np.array([1, -1, 1, -1, 1, -1]),
        [5, 4, 3, 2, 1],
        2 * np.array([1, 1, 1, 1, 1]),
    ]]),
    Matern = dict(kwlist=[dict(nu=v) for v in [0.1, 1, 1.5, 2]]),
    Maternp = dict(kwlist=[dict(p=p) for p in [0, 1, 2]]),
    Log = dict(range=[0, 10]),
    OrnsteinUhlenbeck = dict(range=[0, 3], srange=[0, 10]),
    Periodic = dict(range=[0, 4 * np.pi], kwlist=[dict(outerscale=s) for s in [1, 0.2]]),
    Pink = dict(range=[0, 10], kwlist=[dict(dw=d) for d in [0.1, 1, 10]]),
    Rescaling = dict(skip=True),
    StationaryFracBrownian = dict(kwlist=[dict(H=H) for H in [0.1, 0.5, 0.9]]),
    Taylor = dict(range=[-2, 2]),
    Wendland = dict(range=[0, 2], kwlist=[dict(k=k, alpha=alpha) for k in [0, 2] for alpha in [1, 2]]),
    Wiener = dict(range=[0, 2]),
    WienerIntegral = dict(range=[0, 2]),
    Zeta = dict(x=np.linspace(0, 2, 501), kwlist=[dict(nu=v) for v in [0.1, 1, 1.5, 2.5, 1000]]),
)

fig = plt.figure(num='kernelsref', clear=True)

gen = np.random.default_rng(202206281251)

printoptions = dict(precision=2, threshold=6, edgeitems=2)

# documentation
for kernel in kernels2:
    k = getattr(lgp, kernel)
    sig = str(inspect.signature(k))
    out += f"""\
.. autoclass:: {kernel}{sig}
    :members:
    :class-doc-from: class
"""

    fig.clf()
    ax = fig.subplots()
    
    m = meta.get(kernel, {})
    
    if m.get('skip', False):
        continue

    if 'x' in m:
        x = m['x']
        l = np.min(x)
        r = np.max(x)
    else:
        l, r = m.get('range', [0, 5])
        x = np.linspace(l, r, 500)
    
    legend = m.get('kwlist')
    for kw in m.get('kwlist', [{}]):
        covfun = k(**kw)
        with np.printoptions(**printoptions):
            label = ', '.join(f'{k} = {v}' for k, v in kw.items())
        if issubclass(k, lgp.StationaryKernel):
            acf = covfun(x[0], x)
            ax.plot(x, acf, label=label)
        else:
            cov = covfun(x[None, :], x[:, None])
            dx = (x[1] - x[0]) / 2
            vmin = min(0, np.min(cov))
            im = ax.imshow(cov, aspect='equal', origin='lower', vmin=vmin, extent=(x[0] - dx, x[-1] + dx, x[0] - dx, x[-1] + dx))
            legend = False
            break
    
    if legend:
        ax.legend()
    ax.set_title('Covariance function')
    ax.set_xlabel('x')
    if issubclass(k, lgp.StationaryKernel):
        ax.set_ylabel(f'Cov[f(x), f({l})]')
        b, t = ax.get_ylim()
        dy = 0.1
        lim = max([abs(b), abs(t), 1 + dy])
        ax.set_ylim(-lim, lim)
        ax.axvspan(l, r, -1, 0.5, color='#ddd')
        ax.set_xlim(l, r)
    else:
        ax.set_ylabel("x'")
        fig.colorbar(im, label="Cov[f(x), f(x')]")
    figname = f'kernelsref-{kernel}.png'
    fig.savefig(figname)
    
    out += f"""\
.. image:: {figname}
"""
    
    fig.clf()
    ax = fig.subplots()

    if 'srange' in m:
        l, r = m['srange']
        x = np.linspace(l, r, len(x))
    
    nsamples = 1 if m.get('kwlist', False) else 2
    for kw in m.get('kwlist', [{}]):
        covfun = k(**kw)
        with np.printoptions(**printoptions):
            label = ', '.join(f'{k} = {v}' for k, v in kw.items())
        cov = covfun(x[None, :], x[:, None])
        try:
            dec = lgp._linalg.Chol(cov)
        except np.linalg.LinAlgError:
            dec = lgp._linalg.EigCutFullRank(cov)
        iid = gen.standard_normal((dec.m, nsamples))
        samples = dec.correlate(iid)
        for j, y in enumerate(samples.T):
            ax.plot(x, y, label=None if j else label)
    
    if m.get('kwlist'):
        ax.legend()
    ax.set_title('Samples')
    ax.set_xlabel('x')
    ax.set_ylabel('f(x)')
    figname = f'kernelsref-{kernel}-samples.png'
    fig.savefig(figname)

    out += f"""\
.. image:: {figname}
"""

print(f'writing to {outputfile}...')
with open(outputfile, 'w') as file:
    file.write(out)



# lsqfitgp/docs/examplesref.py
#
# Copyright (c) 2022, 2023, Giacomo Petrillo
#
# This file is part of lsqfitgp.
#
# lsqfitgp is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# lsqfitgp is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with lsqfitgp.  If not, see <http://www.gnu.org/licenses/>.

"""Generate a file with the list of example scripts."""

outputfile = 'examplesref.rst'

import pathlib
import re
import textwrap

examples = pathlib.Path('../examples').glob('*.py')
examples = list(sorted(examples))

shortindex = ''
longindex = ''

for example in examples:
    
    # get script name and url
    name = example.name
    if name == 'runexamples.py':
        continue
    url = f'https://github.com/Gattocrucco/lsqfitgp/blob/master/examples/{name}'
    
    # get description from docstring
    with open(example, 'r') as stream:
        text = stream.read()
    if match := re.search(r'(?s)"""(.+?)"""', text):
        descr = match.group(1)
        descr = ': ' + textwrap.indent(textwrap.dedent(descr), 4 * ' ').strip()
    else:
        descr = ''
        
    # get eventual figure
    base = example.stem
    imgfile = example.parent / 'plot' / f'{base}.png'
    if imgfile.exists():
        image = f"""
    
    .. image:: {imgfile}"""
    else:
        image = ''
    
    # append list item
    shortindex += f'{base}_ '
    longindex += f"""
.. _{base}:

  * `{name} <{url}>`_{descr}{image}
"""

out = f"""\
.. file generated automatically by lsqfitgp/docs/examplesref.py

.. currentmodule:: lsqfitgp

.. _examplesref:

Example scripts
===============

This is an index of the example scripts in the ``examples`` directory in the
repository. The links point to the file preview on github.

Short index
-----------
{shortindex}

Long index
----------
{longindex}
"""

print(f'writing to {outputfile}...')
with open(outputfile, 'w') as file:
    file.write(out)



# lsqfitgp/_fit.py
#
# Copyright (c) 2020, 2022, 2023, Giacomo Petrillo
#
# This file is part of lsqfitgp.
#
# lsqfitgp is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# lsqfitgp is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with lsqfitgp.  If not, see <http://www.gnu.org/licenses/>.

import re
import warnings
import functools
import time
import textwrap
import datetime

import gvar
import jax
from jax import numpy as jnp
import numpy
from scipy import optimize
from jax import tree_util

from . import _GP
from . import _linalg
from . import _patch_jax
from . import _patch_gvar
from . import _array

__all__ = [
    'empbayes_fit',
]

@functools.singledispatch
def token_getter(x):
    return x

@functools.singledispatch
def token_setter(x, token):
    return token

@token_getter.register(jnp.ndarray)
@token_getter.register(numpy.ndarray)
def _(x):
    return x[x.ndim * (0,)] if x.size else x

@token_setter.register(jnp.ndarray)
@token_setter.register(numpy.ndarray)
def _(x, token):
    x = jnp.asarray(x)
    return x.at[x.ndim * (0,)].set(token) if x.size else token


def token_map_leaf(func, x):
    if isinstance(x, (jnp.ndarray, numpy.ndarray)):
        token = token_getter(x)
        @jax.custom_jvp
        def jaxfunc(token):
            return jax.pure_callback(func, token, token, vectorized=True)
        @jaxfunc.defjvp
        def _(p, t):
            return (jaxfunc(*p), *t)
        token = jaxfunc(token)
        return token_setter(x, token)
    else:
        token = token_getter(x)
        token = func(token)
        return token_setter(x, token)

def token_map(func, x):
    return tree_util.tree_map(lambda x: token_map_leaf(func, x), x)

class Logger:
    
    def __init__(self, verbosity=0):
        self._verbosity = verbosity
    
    def _indent(self, text, level=0):
        """ indent a text by provided level or by global current level """
        level = max(0, level + self.loglevel._level)
        prefix = 4 * level * ' '
        return textwrap.indent(text, prefix)
    
    def log(self, message, verbosity=1, *, level=0):
        """ print a message """
        if verbosity > self._verbosity:
            return
        print(self._indent(message, level))

    class _LogLevel:
        """shared context manager to indent messages"""
        
        _level = 0
                
        @classmethod
        def __enter__(cls):
            cls._level += 1
        
        @classmethod
        def __exit__(cls, *_):
            cls._level -= 1
    
    loglevel = _LogLevel()

class empbayes_fit(Logger):

    SEPARATE_JAC = False

    def __init__(
        self,
        hyperprior,
        gpfactory,
        data,
        raises=True,
        minkw={},
        gpfactorykw={},
        jit=False,
        method='gradient',
        initial='priormean',
        verbosity=0,
        covariance='auto',
        fix=None,
        mlkw={},
        forward=False,
    ):
        """
    
        Maximum a posteriori fit.
    
        Maximizes the marginal likelihood of the data with a Gaussian process
        model that depends on hyperparameters, multiplied by a prior on the
        hyperparameters.
    
        Parameters
        ----------
        hyperprior : scalar, array or dictionary of scalars/arrays
            A collection of gvars representing the prior for the
            hyperparameters.
        gpfactory : callable
            A function with signature gpfactory(hyperparams) -> GP object. The
            argument `hyperparams` has the same structure of the empbayes_fit
            argument `hyperprior`. gpfactory must be JAX-friendly, i.e.,
            use jax.numpy and jax.scipy instead of plain numpy/scipy and avoid
            assignments to arrays.
        data : dict, tuple or callable
            Dictionary of data that is passed to `GP.marginal_likelihood` on
            the GP object returned by `gpfactory`. If a tuple, it contains the
            first two arguments to `GP.marginal_likelihood`. If a callable, it
            is called with the same arguments of `gpfactory` and must return
            the argument(s) for `GP.marginal_likelihood`.
        raises : bool, optional
            If True (default), raise an error when the minimization fails.
            Otherwise, use the last point of the minimization as result.
        minkw : dict, optional
            Keyword arguments passed to `scipy.optimize.minimize`, overwrites
            values specified by `empbayes_fit`.
        gpfactorykw : dict, optional
            Keyword arguments passed to `gpfactory`, and also to `data` if it
            is a callable.
        jit : bool
            If True, use jax's jit to compile the minimization target. Default
            False.
        method : str
            Minimization strategy. Options:
        
            'nograd'
                Use a gradient-free method.
            'gradient'
                Use a gradient-only method (default).
            'fisher'
                Use a Newton method with the Fisher information matrix plus
                the hyperprior precision matrix.
        initial : str, scalar, array, dictionary of scalars/arrays
            Starting point for the minimization, matching the format of
            `hyperprior`, or one of the following options:
            
            'priormean'
                Start from the hyperprior mean (default).
            'priorsample'
                Take a random sample from the hyperprior.
        verbosity : int
            An integer indicating how much information is printed on the
            terminal:
    
            0
                No logging (default).
            1
                Minimal report.
            2
                Detailed report.
            3
                Log each iteration.
            4
                More detailed iteration log.
            5
                Print the current parameter values at each iteration.
        covariance : str
            Method to estimate the posterior covariance matrix of the
            hyperparameters:
    
            'fisher'
                Use the Fisher information in the MAP, plus the prior precision,
                as precision matrix.
            'minhess'
                Use the hessian estimate of the minimizer as precision matrix.
            'none'
                Do not estimate the covariance matrix.
            'auto' (default)
                'minhess' if applicable, 'none' otherwise.
        fix : scalar, array or dictionary of scalars/arrays
            A set of booleans, with the same format as `hyperprior`, indicating
            which hyperparameters are kept fixed to their initial value.
            Scalars and arrays are broadcasted to the shape of `hyperprior`.
            If a dictionary, missing keys are treated as False.
        mlkw : dict
            Additional arguments passed to `GP.marginal_likelihood`.
        forward : bool
            Use forward instead of backward (default) derivatives.
    
        Attributes
        ----------
        p : scalar, array or dictionary of scalars/arrays
            A collection of gvars representing the hyperparameters that
            maximize their posterior. These gvars do not track correlations
            with the hyperprior or the data.
        prior : scalar, array or dictionary of scalars/arrays
            A copy of the hyperprior.
        initial : scalar, array or dictionary of scalars/arrays
            Starting point of the minimization, with the same format as `p`.
        fix : scalar, array or dictionary of scalars/arrays
            A set of booleans, with the same format as `p`, indicating which
            parameters were kept fixed to the values in `initial`.
        pmean : scalar, array or dictionary of scalars/arrays
            Mean of `p`.
        pcov : scalar, array or dictionary of scalars/arrays
            Covariance matrix of `p`.
        minresult : scipy.optimize.OptimizeResult
            The result object returned by `scipy.optimize.minimize`.
        minargs : dict
            The arguments passed to `scipy.optimize.minimize`.
        gpfactory : callable
            The `gpfactory` argument.

        Raises
        ------
        RuntimeError
            The minimization failed and `raises` is True.
    
        """

        # TODO parameter float32: bool to use short float type. I think that
        # scipy's optimize may break down with short floats with default
        # options, I hope that changing termination tolerances does the trick.

        Logger.__init__(self, verbosity)
        self.log('**** call lsqfitgp.empbayes_fit ****')
    
        assert callable(gpfactory)
        
        # analyze the hyperprior
        hpinitial, hpunflat = self._parse_hyperprior(hyperprior, initial, fix)
        del hyperprior, initial, fix
        
        # analyze data
        data, cachedargs = self._parse_data(data)

        def make(p):
            
            hp = hpunflat(p)
            gp = gpfactory(hp, **gpfactorykw)
            assert isinstance(gp, _GP.GP)
            
            if cachedargs:
                args = cachedargs
            else:
                args = data(hp, **gpfactorykw)
                if not isinstance(args, tuple):
                    args = (args,)

            return gp, args
        
        def dojit(f):
            return jax.jit(f) if jit else f
        if jit:
            self.log('compile functions with jax jit', 2)
        
        timer = self._Timer()
        firstcall = [None]
        
        @dojit
        def fun(p):
            p = timer.start(p)
            gp, args = make(p)
            decomp, ymean = gp._prior_decomp(*args, covtransf=timer.partial, **mlkw)
            decomp = timer.partial(decomp)

            if firstcall:
                # works under jit since the first call is tracing
                firstcall.pop()
                xdtype = gp._get_x_dtype()
                nd = '?' if xdtype is None else str(_array._nd(xdtype))
                self.log(f'{ymean.size} datapoints, {nd} covariates')
                    
            return timer.partial(1/2 * (
                decomp.logdet()
                + decomp.n * jnp.log(2 * jnp.pi)
                + decomp.quad(ymean)
                + p @ p
            ))

        jactr = jax.jacfwd if forward else jax.jacrev
        self.log(f'{"forward" if forward else "reverse"}-mode autodiff (if used)', 2)
        fun_and_jac = dojit(_patch_jax.value_and_ops(fun, jactr))
        jac = dojit(jactr(fun))

        if not callable(data):
            
            @jax.jacfwd
            @jax.jacfwd
            # can't change inner jac to rev due to jax issue #10994
            # (stop_hessian)
            def fisher(p):
                p = timer.start(p)
                gp, args = make(p)
                decomp, _ = gp._prior_decomp(*args, stop_hessian=True, covtransf=timer.partial, **mlkw)
                return -1/2 * decomp.logdet()
            
            def fisherprec(p):
                f = fisher(p)
                return timer.partial(f + jnp.eye(len(f), dtype=f.dtype))
        
        else:
            
            # Must take into account that the data can depend on the        
            # hyperparameters. The statistical meaning is that the prior mean
            # depends on the hyperparameters.
            
            @functools.partial(jax.jacfwd, has_aux=True)
            # can't change jac to rev due to jax issue #10994 (stop_hessian)
            def grad_logdet_and_aux(p):
                p = timer.start(p)
                gp, args = make(p)
                decomp, ymean = gp._prior_decomp(*args, stop_hessian=True, covtransf=timer.partial, **mlkw)
                return -1/2 * decomp.logdet(), (decomp, ymean)
            
            @functools.partial(jax.jacfwd, has_aux=True)
            def fisher_and_jac_and_aux(p):
                gld, (decomp, ymean) = grad_logdet_and_aux(p)
                return (gld, ymean), decomp
            
            @dojit
            def fisherprec(p):
                (F, J), decomp = fisher_and_jac_and_aux(p)
                result = F + jnp.eye(len(F), dtype=F.dtype) + decomp.quad(J)
                return timer.partial(result)
        
        # wrap functions to count number of calls
        fun = self._CountCalls(fun)
        fun_and_jac = self._CountCalls(fun_and_jac)
        jac = self._CountCalls(jac)
        fisherprec = self._CountCalls(fisherprec)
        
        # prepare minimizer arguments
        if self.SEPARATE_JAC:
            minargs = dict(fun=fun, jac=jac, x0=hpinitial)
        else:
            minargs = dict(fun=fun_and_jac, jac=True, x0=hpinitial)
        if method == 'nograd':
            minargs.update(fun=fun, jac=None, method='nelder-mead')
        elif method == 'gradient':
            minargs.update(method='bfgs')
        elif method == 'fisher':
            minargs.update(hess=fisherprec, method='dogleg')
            # dogleg requires positive definiteness
            # TODO use trust-constr because it has more options?
        else:
            raise KeyError(method)
        self.log(f'method {method!r}', 2)

        # set up callback to time and log iterations
        functions = {
            'fun': fun,
            'jac': jac,
            'fun&jac': fun_and_jac,
            'fisher': fisherprec,
        }
        callback = self._Callback(self, functions, timer, hpunflat)
        minargs.update(callback=callback)

        # check invalid argument before running minimizer
        if not covariance in ('auto', 'fisher', 'minhess', 'none'):
            raise KeyError(covariance)
        
        # add user arguments and minimize
        minargs.update(minkw)
        self.log(f'minimizer method {minargs["method"]!r}', 2)
        start = time.time()
        result = optimize.minimize(**minargs)
        end = time.time()

        # log total timings and function calls
        total = end - start
        times = callback.fmttimes({
            'gp&cov': timer.totals[0],
            'decomp': timer.totals[1],
            'likelihood': timer.totals[2],
            'other': total - sum(timer.totals.values()),
        })
        calls = self._CountCalls.fmtcalls('total', functions)
        self.log('', 4)
        self.log(f'total time: {callback.fmttime(total)}')
        if jit:
            overhead = callback.estimate_firstcall_overhead()
            if overhead is not None:
                self.log(f'estimated compilation time: {callback.fmttime(overhead)}', 2)
        self.log(f'partials: {times}', 2)
        self.log(f'calls: {calls}')
        
        # check the minimization was successful
        if result.success:
            self.log(f'minimization succeeded: {result.message}')
        else:
            msg = f'minimization failed: {result.message}'
            if raises:
                raise RuntimeError(msg)
            elif verbosity == 0:
                warnings.warn(msg)
            else:
                self.log(msg)
        
        # compute posterior covariance of the hyperparameters
        cov = self._posterior_covariance(method, covariance, result, fisherprec)    
        
        uresult = gvar.gvar(result.x, cov)
        
        self.p = gvar.gvar(hpunflat(uresult))
        self.pmean = gvar.mean(self.p)
        self.pcov = gvar.evalcov(self.p)
        self.minresult = result
        self.minargs = minargs
        self.gpfactory = gpfactory

        if verbosity >= 2:
            self.log(_patch_gvar.tabulate_together(
                self.prior, self.p,
                headers=['param', 'prior', 'posterior'],
            )) # TODO replace tabulate_toegether with something more flexible I
            # can use for the callback as well. Maybe import TextMatrix from
            # miscpy.
            # TODO print the transformed parameters
        self.log('**** exit lsqfitgp.empbayes_fit ****')

        # TODO would it be meaningful to add correlation of the fit result with
        # the data and hyperprior?
    
        # TODO add the second order correction. It probably requires more than
        # the gradient and inv_hess, but maybe by getting a little help from
        # marginal_likelihood I can use the least-squares optimized second order
        # correction on the residuals term and invent something for the logdet
        # term.
    
        # TODO it raises very often with "Desired error not necessarily achieved
        # due to precision loss.". I tried doing a forward grad on the logdet
        # but does not fix the problem. I still suspect it's the logdet, maybe
        # the value itself and not the derivative, because as the matrix changes
        # the regularization can change a lot the value of the logdet. How do I
        # stabilize it? => scipy's l-bfgs-b seems to fail the linear search less
        # often
        
        # TODO compute the logGBF for the whole fit (see the gpbart code)
        
        # TODO empbayes_fit(autoeps=True) tries to double epsabs until the
        # minimization succedes, with some maximum number of tries.
        # autoeps=dict(maxrepeat=5, increaseby=2, initial=1e-16,
        # startfromzero=True) allows to configure the algorithm.

        # TODO empbayes_fit(maxiter=100) sets the maximum number of minimization
        # iterations. maxiter=dict(iter=100, calls=200, callsperiter=10) allows
        # to configure it more finely. The calls limits are cumulative on all
        # functions (need to make a class counter in _CountCalls), I can
        # probably implement them by returning nan when the limit is surpassed,
        # I hope the minimizer stops immediately on nan (test this).

        # TODO can I approximate the hessian with only function values and no
        # gradient, i.e., when using nelder-mead? => See Hare (2022), although I
        # would not know how to apply it properly to the optimization history.
        # Somehow I need to keep only the "last" iterations.

        # TODO is there a better algorithm than lbfgs for inaccurate functions?
        # consider SC-BFGS (https://github.com/frankecurtis/SCBFGS). See Basak 
        # (2022). And NonOpt (https://github.com/frankecurtis/NonOpt).

        # TODO can I estimate the error on the likelihood with the matrices? It
        # requires the condition number. Basak (2022) gives wide bounds. I could
        # try an upper bound and see how it compares to the true error, assuming
        # that the matrix was as ill-conditioned as possible, i.e., use eps as
        # the lowest eigenvalue, and gershgorin as the highest one.

        # TODO look into jaxopt: it has improved a lot since the last time I saw
        # it. In particular, it implements l-bfgs and has a "do not stop on
        # failed line search" option. And it probably supports float32.

        # TODO I could provide the option to use BFGS to estimate the covariance
        # independently of the minimizer, using the last n iterations (not all
        # the evaluations, just the iterations, to skip difficult line
        # searches.) covariance='bfgs(10)'. I still need, in any case, an
        # automatic policy of this sort for covariance='hessinv' when the linear
        # search fails.

        # TODO reimplement the timing system with host_callback.id_tap. It
        # should preserve the order because id_tap takes inputs and outputs. I
        # must take care to make all callbacks happen at runtime instead of
        # having some of them at compile time. I tried once but failed.
        # Currently host_callback is experimental, maybe wait until it isn't. =>
        # I think it fails because it's asynchronous and there is only one
        # device. Maybe host_callback.call would work?

    class _CountCalls:
        """ wrap a callable to count calls """
        
        def __init__(self, func):
            self._func = func
            self._total = 0
            self._partial = 0
            functools.update_wrapper(self, func)
        
        def __call__(self, *args, **kw):
            self._total += 1
            self._partial += 1
            return self._func(*args, **kw)
        
        def partial(self):
            """ return the partial counter and reset it """
            result = self._partial
            self._partial = 0
            return result
        
        def total(self):
            """ return the total number of calls """
            return self._total

        @staticmethod
        def fmtcalls(method, functions):
            """
            format summary of number of calls
            method : str
            functions: dict[str, _CountCalls]
            """
            def counts():
                for name, func in functions.items():
                    if count := getattr(func, method)():
                        yield f'{name} {count}'
            return ', '.join(counts())

    class _Timer:
        """ object to time likelihood computations """

        def __init__(self):
            self.totals = {}
            self.partials = {}
            self._last_start = False

        def start(self, token):
            return token_map(self._start, token)

        def _start(self, token):
            self.stamp = time.time()
            self.counter = 0
            assert not self._last_start
            self._last_start = True
            return token

        def reset(self):
            self.partials = {}

        def partial(self, token):
            return token_map(self._partial, token)

        def _partial(self, token):
            now = time.time()
            delta = now - self.stamp
            self.partials[self.counter] = self.partials.get(self.counter, 0) + delta
            self.totals[self.counter] = self.totals.get(self.counter, 0) + delta
            self.stamp = now
            self.counter += 1
            self._last_start = False
            return token
    
    def _parse_hyperprior(self, hyperprior, initial, fix):
        
        # check fix against hyperprior and fill missing values
        hyperprior = self._copyasarrayorbufferdict(hyperprior)
        self._check_no_redundant_keys(hyperprior)
        fix = self._parse_fix(hyperprior, fix)
        flatfix = self._flatview(fix)

        # extract distribution of free hyperparameters
        flathp = self._flatview(hyperprior)
        freehp = flathp[~flatfix]
        mean = gvar.mean(freehp)
        cov = gvar.evalcov(freehp) # TODO use evalcov_blocks
        dec = _linalg.EigCutFullRank(cov)
        assert dec.n == freehp.size
        self.log(f'{freehp.size}/{flathp.size} free hyperparameters', 2)
        
        # determine starting point for minimization
        initial = self._parse_initial(hyperprior, initial, dec)
        flatinitial = self._flatview(initial)
        x0 = dec.decorrelate(flatinitial[~flatfix] - mean)
        # TODO for initial = 'priormean', x0 is zero, skip decorrelate
        # for initial = 'priorsample', x0 is iid normal, but I have to sync
        # it with the user-exposed unflattened initial in _parse_initial
        
        # make function to correlate, add fixed values, and reshape to original
        # format
        fixed_indices, = jnp.nonzero(flatfix)
        unfixed_indices, = jnp.nonzero(~flatfix)
        fixed_values = jnp.asarray(flatinitial[flatfix])
        def unflat(x):
            assert x.ndim == 1
            if x.dtype == object:
                jac, indices = _patch_gvar.jacobian(x)
                xmean = mean + dec.correlate(gvar.mean(x))
                xjac = dec.correlate(jac)
                x = _patch_gvar.from_jacobian(xmean, xjac, indices)
                y = numpy.empty(flatfix.size, x.dtype)
                y[unfixed_indices] = x
                y[fixed_indices] = fixed_values
            else:
                x = mean + dec.correlate(x)
                y = jnp.empty(flatfix.size, x.dtype)
                y = y.at[unfixed_indices].set(x)
                y = y.at[fixed_indices].set(fixed_values)
            return self._unflatview(y, hyperprior)
        
        self.prior = hyperprior
        return x0, unflat
    
    @staticmethod
    def _check_no_redundant_keys(hyperprior):
        if not hasattr(hyperprior, 'keys'):
            return
        for k in hyperprior:
            m = hyperprior.extension_pattern.match(k)
            if m and m.group(1) in hyperprior.invfcn:
                altk = m.group(2)
                if altk in hyperprior:
                    raise ValueError(f'duplicate keys {altk!r} and {k!r} in hyperprior')

    def _parse_fix(self, hyperprior, fix):
        
        if fix is None:
            if hasattr(hyperprior, 'keys'):
                fix = gvar.BufferDict(hyperprior, buf=numpy.zeros(hyperprior.size, bool))
            else:
                fix = numpy.zeros(hyperprior.shape, bool)
        else:
            fix = self._copyasarrayorbufferdict(fix)
            if hasattr(fix, 'keys'):
                assert hasattr(hyperprior, 'keys'), 'fix is dictionary but hyperprior is array'
                assert all(hyperprior.has_dictkey(k) for k in fix), 'some keys in fix are missing in hyperprior'
                newfix = {}
                for k, v in hyperprior.items():
                    key = None
                    m = hyperprior.extension_pattern.match(k)
                    if m and m.group(1) in hyperprior.invfcn:
                        altk = m.group(2)
                        if altk in fix:
                            assert k not in fix, f'duplicate keys {k!r} and {altk!r} in fix'
                            key = altk
                    if key is None and k in fix:
                        key = k
                    if key is None:
                        elem = numpy.zeros(v.shape, bool)
                    else:
                        elem = numpy.broadcast_to(fix[key], v.shape)
                    newfix[k] = elem
                fix = gvar.BufferDict(newfix, dtype=bool)
            else:
                assert not hasattr(hyperprior, 'keys'), 'fix is array but hyperprior is dictionary'
                fix = numpy.broadcast_to(fix, hyperprior.shape).astype(bool)
        
        self.fix = fix
        return fix
    
    def _parse_initial(self, hyperprior, initial, dec):
        
        if not isinstance(initial, str):
            self.log('start from provided point', 2)
            initial = self._copyasarrayorbufferdict(initial)
            if hasattr(hyperprior, 'keys'):
                assert hasattr(initial, 'keys'), 'hyperprior is dictionary but initial is array'
                assert set(hyperprior.keys()) == set(initial.keys())
                assert all(hyperprior[k].shape == initial[k].shape for k in hyperprior)
            else:
                assert not hasattr(initial, 'keys'), 'hyperprior is array but initial is dictionary'
                assert hyperprior.shape == initial.shape
        
        elif initial == 'priormean':
            self.log('start from prior mean', 2)
            initial = gvar.mean(hyperprior)
        
        elif initial == 'priorsample':
            self.log('start from a random sample from the prior', 2)
            if dec.n < hyperprior.size:
                flathp = self._flatview(hyperprior)
                cov = gvar.evalcov(flathp) # TODO use evalcov_blocks
                fulldec = _linalg.EigCutFullRank(cov)
            else:
                fulldec = dec
            iid = numpy.random.randn(fulldec.m)
            flatinitial = numpy.asarray(fulldec.correlate(iid))
            initial = self._unflatview(flatinitial, hyperprior)
        
        else:
            raise KeyError(initial)
        
        self.initial = initial
        return initial
    
    def _parse_data(self, data):
        
        if isinstance(data, tuple) and len(data) == 1:
            data, = data

        if callable(data):
            self.log('data is callable', 2)
            cachedargs = None
        elif isinstance(data, tuple):
            self.log('data errors provided separately', 2)
            assert len(data) == 2
            cachedargs = data
        elif (gdata := self._copyasarrayorbufferdict(data)).dtype == object:
            self.log('data has errors as gvars', 2)
            data = gvar.gvar(gdata)
            # convert to gvar because non-gvars in the array would upset
            # gvar.mean and gvar.evalcov
            cachedargs = (gvar.mean(data), gvar.evalcov(data))
        else:
            self.log('data has no errors', 2)
            cachedargs = (data,)
        
        return data, cachedargs

    def _posterior_covariance(self, method, covariance, minimizer_result, fisher_func):
        
        if covariance == 'auto':
            if hasattr(minimizer_result, 'hess_inv') or hasattr(minimizer_result, 'hess'):
                covariance = 'minhess'
            else:
                covariance = 'none'

        if covariance == 'fisher':
            self.log('use fisher plus prior precision as precision', 2)
            if method == 'fisher':
                prec = minimizer_result.hess
            else:
                prec = fisher_func(minimizer_result.x)
            cov = _linalg.EigCutFullRank(prec).inv()

        elif covariance == 'minhess':
            if hasattr(minimizer_result, 'hess_inv'):
                hessinv = minimizer_result.hess_inv
                if isinstance(hessinv, optimize.LbfgsInvHessProduct):
                    self.log(f'convert LBFGS({hessinv.n_corrs}) hessian inverse to BFGS as covariance', 2)
                    cov = self._invhess_lbfgs_to_bfgs(hessinv)
                    # TODO this still gives a too wide cov when the minimization
                    # terminates due to bad linear search, is it because of
                    # dropped updates? This is currently keeping me from setting
                    # l-bfgs-b as default minimization method.
                elif isinstance(hessinv, numpy.ndarray):
                    self.log('use minimizer estimate of inverse hessian as covariance', 2)
                    cov = hessinv
            elif hasattr(minimizer_result, 'hess'):
                self.log('use minimizer hessian as precision', 2)
                cov = _linalg.EigCutFullRank(minimizer_result.hess).inv()
            else:
                raise RuntimeError('the minimizer did not return an estimate of the hessian')

        elif covariance == 'none':
            cov = numpy.full(minimizer_result.x.size, numpy.nan)

        else:
            raise KeyError(covariance)

        return cov

    @staticmethod
    def _invhess_lbfgs_to_bfgs(lbfgs):
        bfgs = optimize.BFGS()
        bfgs.initialize(lbfgs.shape[0], 'inv_hess')
        for i in range(lbfgs.n_corrs):
            bfgs.update(lbfgs.sk[i], lbfgs.yk[i])
        return bfgs.get_matrix()

    class _Callback:
        """ Iteration callback for scipy.optimize.minimize """
        
        def __init__(self, this, functions, timer, unflat):
            self.it = 0
            self.stamp = time.time()
            self.this = this
            self.functions = functions
            self.timer = timer
            self.unflat = unflat
            self.tail_overhead = 0
            self.tail_overhead_iter = 0

        def __call__(self, p):
            self.it += 1
            now = time.time()
            duration = now - self.stamp
            
            worktime = sum(self.timer.partials.values())
            if worktime:
                overhead = duration - worktime
                assert overhead >= 0, (duration, worktime)
                if self.it == 1:
                    self.first_overhead = overhead
                else:
                    self.tail_overhead_iter += 1
                    self.tail_overhead += overhead

            if self.this._verbosity >= 3:
                calls = self.this._CountCalls.fmtcalls('partial', self.functions)

            if self.this._verbosity >= 4:
                tot = self.fmttime(duration)
                if self.timer.partials:
                    times = {
                        'gp&cov': self.timer.partials[0],
                        'dec': self.timer.partials[1],
                        'like': self.timer.partials[2],
                        'other': duration - sum(self.timer.partials.values()),
                    }
                    times = self.fmttimes(times)
                else:
                    times = 'n/d'
                self.this.log(f'\niteration {self.it}', 4)
                with self.this.loglevel:
                    self.this.log(f'total time: {tot}', 4)
                    self.this.log(f'partial: {times}', 4)
                    self.this.log(f'calls: {calls}', 4)

                if self.this._verbosity >= 5:
                    nicep = self.unflat(p)
                    nicep = self.this._copyasarrayorbufferdict(nicep)
                    with self.this.loglevel:
                        self.this.log(f'parameters = {nicep}', 5)
                    # TODO write a method to format the parameters nicely. => use
                    # gvar.tabulate? => nope, need actual gvars

            elif self.this._verbosity >= 3:
                times = self.fmttime(duration)
                self.this.log(f'iter {self.it}, time: {times}, calls: {calls}', 3)

            self.stamp = now
            self.timer.reset()

        pattern = re.compile(
            r'((\d+) days, )?(\d{1,2}):(\d\d):(\d\d(\.\d{6})?)')

        @classmethod
        def fmttime(cls, seconds):
            td = datetime.timedelta(seconds=seconds)
            m = cls.pattern.fullmatch(str(td))
            _, day, hour, minute, second, _ = m.groups()
            hour = int(hour)
            minute = int(minute)
            second = float(second)
            if day:
                return f'{day.lstrip("0")}d{hour:02d}h'
            elif hour:
                return f'{hour}h{minute:02d}m'
            elif minute:
                return f'{minute}m{second:02.0f}s'
            elif second >= 0.0995:
                return f'{second:#.2g}'.rstrip('.') + 's'
            elif second >= 0.0000995:
                return f'{second * 1e3:#.2g}'.rstrip('.') + 'ms'
            else:
                return f'{second * 1e6:.0f}μs'

        @classmethod
        def fmttimes(cls, times):
            """ times = dict label -> seconds """
            return ', '.join(f'{k} {cls.fmttime(v)}' for k, v in times.items())

        def estimate_firstcall_overhead(self):
            if self.tail_overhead_iter and hasattr(self, 'first_overhead'):
                typical_overhead = self.tail_overhead / self.tail_overhead_iter
                return self.first_overhead - typical_overhead

    @staticmethod
    def _copyasarrayorbufferdict(x):
        if hasattr(x, 'keys'):
            return gvar.BufferDict(x)
        else:
            return numpy.array(x)

    @staticmethod
    def _flatview(x):
        if hasattr(x, 'reshape'):
            return x.reshape(-1)
        elif hasattr(x, 'buf'):
            return x.buf
        else: # pragma: no cover
            raise NotImplementedError

    @staticmethod
    def _unflatview(x, original):
        if isinstance(original, numpy.ndarray):
            # TODO is this never applied to jax arrays?
            out = x.reshape(original.shape)
            # if not out.shape:
            #     try:
            #         out = out.item()
            #     except jax.errors.ConcretizationTypeError:
            #         pass
            return out
        elif isinstance(original, gvar.BufferDict):
            # normally I would do BufferDict(original, buf=x) but it does not
            # work with JAX tracers
            b = gvar.BufferDict(original)
            b._extension = {}
            b._buf = x
            # b.buf = x does not work because BufferDict checks that the
            # array is a numpy array
            # TODO maybe make a feature request to gvar to accept array_like
            # buf
            return b
        else: # pragma: no cover
            raise NotImplementedError



import functools

import jax
from jax import numpy as jnp
from jax import core
from jax.interpreters import ad
from jax.interpreters import batching
from jax import abstract_arrays
from scipy import special

def makejaxufunc(ufunc, *derivs):
    prim = core.Primitive(ufunc.__name__)
    
    @functools.wraps(ufunc)
    def func(*args):
        return prim.bind(*args)

    @prim.def_impl
    def impl(*args):
        return ufunc(*args)

    @prim.def_abstract_eval
    def abstract_eval(*args):
        return x

    jvps = (
        None if d is None
        else lambda g, *args: d(*args) * g
        for d in derivs
    )
    ad.defjvp(prim, *jvps)

    batching.defvectorized(prim)
    
    return func

j0 = makejaxufunc(special.j0, lambda x: -j1(x))
j1 = makejaxufunc(special.j1, lambda x: (j0(x) - jn(2, x)) / 2.0)
jn = makejaxufunc(special.jn, None, lambda n, x: (jn(n - 1, x) - jn(n + 1, x)) / 2.0)



import numpy as np
import numba

default_bart_params = dict(m=200, alpha=0.95, beta=2, k=2, nu=3, q=0.9)

@numba.jit(nopython=True, cache=True, fastmath=True)
def bart_correlation_maxd_recursive(nminus, n0, nplus, d, alpha, beta, maxd):
    sump = 0.0
    p = len(nminus)
    for i in range(p):
        nminusi = nminus[i]
        n0i = n0[i]
        nplusi = nplus[i]
        
        if n0i == 0:
            sump += 1
        elif d >= maxd:
            sump += (nminusi + nplusi) / (nminusi + n0i + nplusi)
        else:
            sumn = 0
            for k in range(nminusi):
                nminus[i] = k
                sumn += bart_correlation_maxd_recursive(nminus, n0, nplus, d + 1, alpha, beta, maxd)
            nminus[i] = nminusi
            for k in range(nplusi):
                nplus[i] = k
                sumn += bart_correlation_maxd_recursive(nminus, n0, nplus, d + 1, alpha, beta, maxd)
            nplus[i] = nplusi
            sump += sumn / (nminusi + n0i + nplusi)
    
    pnt = alpha * (1 + d) ** -beta
    return 1 - pnt * (1 - sump / p)

@numba.guvectorize('(i8[:], i8[:], i8[:], f8, f8, i8, f8[:])', '(p),(p),(p),(),(),()->()', nopython=True)
def bart_correlation_maxd_vectorized(nminus, n0, nplus, alpha, beta, maxd, out):
    assert np.all(nminus >= 0)
    assert np.all(n0 >= 0)
    assert np.all(nplus >= 0)
    assert maxd >= 0
    if maxd == 0:
        out[0] = 1
    else:
        out[0] = bart_correlation_maxd_recursive(nminus, n0, nplus, 0, alpha, beta, maxd - 1)

def bart_correlation_maxd(splitsbefore, splitsbetween, splitsafter, alpha, beta, maxd):
    """
    Compute the BART prior correlation between two points.
    
    The correlation is computed approximately by limiting the maximum depth
    of the trees. Limiting trees to depth 1 is equivalent to setting beta to
    infinity.
    
    The function is fully vectorized.
    
    Parameters
    ----------
    splitsbefore : int (p,) array
        The number of splitting points less than the two points, separately
        along each coordinate.
    splitsbetween : int (p,) array
        The number of splitting points between the two points, separately along
        each coordinate.
    splitsafter : int (p,) array
        The number of splitting points greater than the two points, separately
        along each coordinate.
    alpha, beta : scalar
        The hyperparameters of the branching probability.
    maxd : int
        The maximum depth of the trees. The root has depth zero.
    
    Return
    ------
    corr : scalar
        The prior correlation.
    """
    return bart_correlation_maxd_vectorized(splitsbefore, splitsbetween, splitsafter, alpha, beta, maxd)



from scipy import stats, linalg
import numpy as np
from matplotlib import pyplot as plt

def gen1(n, size=1):
    return stats.special_ortho_group.rvs(n, size)

def gen2(n, size=()):
    if not hasattr(size, '__len__'):
        size = (size,)
    a = np.random.randn(*size, n, n)
    q, r = np.linalg.qr(a)
    # q *= np.sign(np.diagonal(r, 0, -2, -1))[..., None, :]
    q *= 2 * np.random.randint(2, size=size + (1, n)) - 1
    return q

def gen3(n, size=()):
    if not hasattr(size, '__len__'):
        size = (size,)
    a = np.random.randn(*size, n, n)
    at = np.swapaxes(a, -2, -1)
    s = a + at
    w, v = np.linalg.eigh(s)
    return v
    
def gen4(n, size=()):
    pass
    # TODO try to use the h, tau vectors returned by the mode='raw' option
    # of np.linalg.qr, following the last algorithm in Mezzadri (2006).

samples = [
    np.angle(np.linalg.eigvals(gen(3, 10000))).flat
    for gen in [gen1, gen2]
]

fig, ax = plt.subplots(num='randso', clear=True)

ax.hist(samples, bins='auto', histtype='step')

plt.show()



from scipy import linalg
import numpy as np

def assert_close_matrices(actual, desired, *, rtol=0, atol=0):
    actual = np.asarray(actual)
    desired = np.asarray(desired)
    assert actual.shape == desired.shape
    if actual.size == 0:
        return
    actual = np.atleast_1d(actual)
    desired = np.atleast_1d(desired)
    
    dnorm = linalg.norm(desired, 2)
    adnorm = linalg.norm(actual - desired, 2)
    ratio = adnorm / dnorm if dnorm else np.nan
    msg = f"""\
matrices actual and desired are not close in 2-norm
norm(desired) = {dnorm:.2g}
norm(actual - desired) = {adnorm:.2g}  (atol = {atol:.2g})
ratio = {ratio:.2g}  (rtol = {rtol:.2g})"""
    assert adnorm <= atol + rtol * dnorm, msg

def eval_matrix_difference(expr1, expr2):
    actual = np.asarray(eval(expr1))
    desired = np.asarray(eval(expr2))
    assert actual.shape == desired.shape
    actual = np.atleast_1d(actual)
    desired = np.atleast_1d(desired)
    dnorm = linalg.norm(desired, 2)
    adnorm = linalg.norm(actual - desired, 2)
    ratio = adnorm / dnorm if dnorm else np.nan
    print(f'norm({expr1} - {expr2}) / norm({expr2}) = {ratio:.2g}')

def pinv_svd(a):
    u, s, vh = linalg.svd(a, full_matrices=False)
    eps = np.max(s) * np.finfo(s.dtype).eps * max(a.shape)
    si = np.where(s > eps, 1 / np.where(s, s, 1), 0)
    return (vh.T * si) @ u.T
    
def pinv_diag(a):
    n, m = a.shape
    transpose = n < m
    if transpose:
        a = a.T
    # a+ = (at a)+ at
    # at a = v w vt
    w, v = linalg.eigh(a.T @ a)
    eps = np.max(np.abs(w)) * np.finfo(w.dtype).eps * len(w)
    wi = np.where(w > eps, 1 / np.where(w, w, 1), 0)
    ai = np.linalg.multi_dot([v * wi, v.T, a.T])
    if transpose:
        ai = ai.T
    return ai

def pinv_chol(a, *, order=1, epsfac=1024):
    n, m = a.shape
    transpose = n < m
    if transpose:
        a = a.T
    
    # a+ = sum n=1^∞  (at a + eI)^-n e^(n-1) at
    # at a + eI = l lt
    
    aa = a.T @ a
    eps = np.max(np.sum(np.abs(aa), axis=0)) * np.finfo(aa.dtype).eps * len(aa)
    eps *= epsfac
    aa[np.diag_indices_from(aa)] += eps
    l = linalg.cholesky(aa, lower=True)
    
    term = a.T
    ai = 0
    for i in range(order):
        l_1term = linalg.solve_triangular(l, term, lower=True)
        term = linalg.solve_triangular(l.T, l_1term, lower=False)
        ai += eps ** i * term
    
    if transpose:
        ai = ai.T
    return ai

n = 1000
m = 100
gen = np.random.default_rng(202301301424)
a = gen.standard_normal((n, m))
a[:, :50] = a[:, :1]

ai = pinv_svd(a)
ai2 = pinv_diag(a)
ai3 = pinv_chol(a, order=2, epsfac=128000)

eval_matrix_difference('ai2', 'ai')
eval_matrix_difference('ai3', 'ai')
print()
eval_matrix_difference('a @ ai @ a', 'a')
eval_matrix_difference('a @ ai2 @ a', 'a')
eval_matrix_difference('a @ ai3 @ a', 'a')
print()
eval_matrix_difference('ai @ a @ ai', 'ai')
eval_matrix_difference('ai2 @ a @ ai2', 'ai2')
eval_matrix_difference('ai3 @ a @ ai3', 'ai3')
print()
eval_matrix_difference('a @ ai', '(a @ ai).T')
eval_matrix_difference('a @ ai2', '(a @ ai2).T')
eval_matrix_difference('a @ ai3', '(a @ ai3).T')
print()
eval_matrix_difference('ai @ a', '(ai @ a).T')
eval_matrix_difference('ai2 @ a', '(ai2 @ a).T')
eval_matrix_difference('ai3 @ a', '(ai3 @ a).T')



import numpy as np
from scipy import special
from matplotlib import pyplot as plt

import lsqfitgp as lgp

fig, ax = plt.subplots(num='expn', clear=True)

nt = 20
eps = np.finfo(float).eps
d = 3
nrange = np.arange(2, 2 + 1 + 10 * d, d)[::-1]

x = np.logspace(0, 4, 10000)
for n in nrange:
    y1 = lgp._patch_jax.expn_imag_smallx(n, x).real
    y2 = lgp._patch_jax.expn_asymp(n, -1j * x, nt).real
    line, = ax.plot(x, np.abs(y1 - y2), label=f'n={n}')
    color = line.get_color()
    # knee = (1/eps) ** (1 / (2 * (n - 1)))
    # knee = (special.poch(n, nt) / eps) ** (1 / (n + nt - 1))
    kneex = (special.gamma(n + nt) / eps) ** (1 / (n + nt - 1))
    kneey = eps * kneex ** (n - 2) / special.gamma(n)
    kw = dict(color=color, linestyle='--')
    ax.axvline(kneex, 0, 1, **kw)
    ax.axhline(kneey, 0, 1, **kw)

ax.set_ylabel('|E_n direct - E_n asymp|')
ax.set_yscale('log')
ax.set_xscale('log')
ax.legend()

fig.show()


fig, ax = plt.subplots(num='expn2', clear=True)

nts = np.linspace(0, 100, 1000)
for n in nrange:
    kneex = (special.gamma(n + nts) / eps) ** (1 / (n + nts - 1))
    kneey = eps * kneex ** (n - 2) / special.gamma(n)
    ax.plot(nts, kneey, label=f'n={n}')

ax.legend()
ax.set_yscale('log')
ax.set_xlabel('number of terms in asymp series')
ax.set_ylabel('error at knee')

fig.show()


fig, ax = plt.subplots(num='expn3', clear=True)

nt = 20
n = np.arange(2, 21)
kneex = (special.gamma(n + nt) / eps) ** (1 / (n + nt - 1))
ax.plot(n, kneex)

ax.set_xlabel('n')
ax.set_ylabel('knee x')

fig.show()

# TODO find the minimum of kneey(nt) for float32 and float64, plot w.r.t. n,
# round to nearest integer, tabulate up to reasonable n (guess 30-40)
